"""Module with functions helpers to  create gasout for CFD solvers"""

import yaml
import copy
import h5py
import numpy as np
from h5cross import hdfdict
import shutil
import pkg_resources
from ms_thermo.state import State


HEADLOG = """
   ### Gasout tool from ms_thermo package ###

    This is a demonstrator!
    It has not been thoroughy tested yet!

"""

__all__ = [
    "gasout_with_input",
    "gasout_dump_default_input",
    "alter_state",
    "spherical_tanh_mask",
    "directional_linear_mask",
    "fraction_z_mask",
    "load_mesh_and_solution",
    "save_data_for_avbp",
    "gasout_tool"
]


def alter_state(state, alpha, temp_new=None, press_new=None, y_new=None, verbose=False):
    """
    Apply gasout alterations to a State.

    :param state: State before alterations
    :type state: State
    :param alpha: Alteration mask with values from 0 (no alteration) to 1 (full alteration)
    :type alpha: ndarray or scalar
    :param temp_new: New temperature to apply, defaults to None
    :type temp_new: ndarray or scalar, optional
    :param press_new: New temperature to apply, defaults to None
    :type press_new: ndarray or scalar, optional
    :param y_new: New mass fractions to apply, defaults to None
    :type y_new: dict[str, ndarray or scalar], optional
    :param verbose: Verbosity, defaults to False
    :type verbose: bool, optional

    :returns: log, debug log
    :rtype: str
    """
    log = str()

    # Order may be important? According to commit d36c7348, Yk must be updated before T
    if y_new is not None:
        y_mix = dict()
        state_species = list(state.mass_fracs.keys())

        for spec in y_new:
            if spec not in state_species:
                msgerr = "\n\n Species mismatch:"
                msgerr += "\n" + str(spec) + " is not part of the mixture:\n"
                msgerr += "/".join(state_species)
                msgerr += "\nCheck your input file pretty please..."
                raise RuntimeWarning(msgerr)

        for spec in state.mass_fracs:
            value = 0
            if spec in y_new:
                value = y_new[spec]

            y_mix[spec] = state.mass_fracs[spec] + alpha * (
                value - state.mass_fracs[spec]
            )
        state.mass_fracs = y_mix

    if temp_new is not None:
        temp = np.array(state.temperature)
        state.temperature = temp + alpha * (temp_new - temp)

    if press_new is not None:
        state.pressure = state.pressure + alpha * (press_new - state.pressure)

    if verbose:
        footprint = 100.0 * np.sum(alpha) / alpha.shape[0]
        log += f"Footprint : {footprint: .3f} %"
    return log


def spherical_tanh_mask(coor, center, radius, delta):
    """
    Define a spherical mask. 0 inside the sphere, 1 outside, with a tanh transition.

    :param coor: Array of spatial coordinates (shape (n, ndim))
    :type coor: ndarray
    :param center: Array of sphere center coordinates (shape (ndim,))
    :type center: list, tuple or ndarray
    :param radius: Radius of the sphere
    :type radius: float
    :param delta: Transition thickness at the edge of the sphere
    :type delta: float

    :returns: alpha, alteration mask
    :rtype: ndarray
    """
    center = np.array(center)
    center_ndim = center.shape[0]
    coor_ndim = coor.shape[-1]
    if center_ndim != coor_ndim:
        msg_err = "\n\nDimension mismatch in the spherical center"
        msg_err += "\nCenter is " + str(center_ndim) + "D"
        msg_err += "\nCoords are " + str(coor_ndim) + "D"
        msg_err += "\nCheck your input file pretty please..."
        raise RuntimeWarning(msg_err)

    r_coor = np.linalg.norm(coor - center, axis=1)
    alpha = 0.5 * (1.0 - np.tanh((r_coor - radius) / (delta / 2)))
    return alpha


def directional_linear_mask(
    coor, axis, transition_start, transition_end
):
    """
    Define a directional mask aligned with the `axis` coordinate axis.

    If `transition_end` > `transition_start`, the mask is 0 before `transition_start` and
    1 after `transition_end.
    If `transition_start` > `transition_end`, the mask is 1 before `transition_end` and
    0 after `transition_start.
    A linear transition is imposed between `transition_start` and `transition_end`.

    :param coor: Array of spatial coordinates (shape (n, ndim))
    :type coor: ndarray
    :param axis: Coordinate axis of the mask (0 x, 1 y, 2 z)
    :type axis: int
    :param transition_start: Start point of the linear mask
    :type transition_start: float
    :param transition_end: End point of the linear mask
    :type transition_end: float

    :returns: alpha, alteration mask
    :rtype: ndarray
    """
    direction = coor[:, axis]

    alpha = (direction - transition_start) / (transition_end - transition_start)
    alpha = np.clip(alpha, 0.0, 1.0)

    return alpha


def fraction_z_mask(
    state,
    specfuel,
    zmin,
    zmax,
    fuel_mass_fracs=None,
    oxyd_mass_fracs=None,
    atom_ref='C',
    verbose=False,
):
    """
    Compute a mask based on the mixture fraction Z.

    The mask is 1 between `zmin` and `zmax` and 0 outside.

    :param state: ms_thermo State object
    :type state: State
    :param specfuel: Fuel species name
    :type specfuel: str
    :param zmin: mask disabled (0) below this value
    :type zmin: float
    :param zmax: mask disabled (0) over this value
    :type zmax: float
    :param fuel_mass_fracs: Fuel mass fractions, defaults to composition at peak fuel concentration
    :type fuel_mass_fracs: dict, optional
    :param oxyd_mass_fracs: Oxydizer mass fractions, defaults to air
    :type oxyd_mass_fracs: dict, optional
    :param atom_ref: Reference atom, defaults to C
    :type atom_ref: str, optional
    :param verbose: Verbosity, defaults to False
    :type verbose: bool, optional

    :returns: alpha, alteration mask
    :rtype: ndarray
    """
    z_frac = state.compute_z_frac(
        specfuel,
        fuel_mass_fracs,
        oxyd_mass_fracs,
        atom_ref,
        verbose,
    )

    z_mid = 0.5 * (zmax + zmin)
    z_gap = 0.5 * (zmax - zmin)
    alpha = np.where(abs(z_frac - z_mid) < z_gap, 1.0, 0)
    return alpha


def gasout_with_input(coor, state, in_nob):
    """
    Update a State with gasout actions.

    :param coor: Array of spatial coordinates (shape (n, ndim))
    :type coor: ndarray
    :param state: ms_thermo State object
    :type state: State
    :param in_nob: Contents of the input file
    :type in_nob: dict

    :returns: Tuple containing the updated State and a log
    :rtype: tuple
    """

    log = str(HEADLOG)
    log += "\nInitial state"
    log += "\n ==========="
    log += "\n" + state.__repr__()

    for i, action in enumerate(in_nob["actions"]):

        log += "\n\nAction " + str(i) + ":" + action["type"]

        if action["type"] == "directional_linear_mask":
            reqd_keys = [
                "direction",
                "transition_start",
                "transition_stop",
            ]
            check = _check_action_params(reqd_keys, list(action.keys()))
            if check != "":
                raise RuntimeWarning(log + check)

            if action["direction"] == "x":
                axis = 0
            elif action["direction"] == "y":
                axis = 1
            elif action["direction"] == "z":
                axis = 2
            else:
                msgerr = "\n INPUT ERROR : " + "Direction must be one of x, y or z"
                raise RuntimeWarning(log + msgerr)

            alpha = directional_linear_mask(
                coor,
                axis=axis,
                transition_start=action["transition_start"],
                transition_stop=action["transition_stop"],
            )

        elif action["type"] == "spherical_tanh_mask":
            reqd_keys = [
                "center",
                "radius",
                "delta",
            ]
            check = _check_action_params(reqd_keys, list(action.keys()))
            if check != "":
                raise RuntimeWarning(log + check)

            alpha = spherical_tanh_mask(
                coor,
                center=action["center"],
                radius=action["radius"],
                delta=action["delta"],
            )

        elif action["type"] == "fraction_z_mask":
            reqd_keys = [
                "specfuel",
                "atom_ref",
                "oxyd_mass_fracs",
                "fuel_mass_fracs",
                "zmax",
                "zmin",
            ]

            check = _check_action_params(reqd_keys, list(action.keys()))
            if check != "":
                raise RuntimeWarning(log + check)

            alpha = fraction_z_mask(
                state,
                action["specfuel"],
                zmin=action["zmin"],
                zmax=action["zmax"],
                fuel_mass_fracs=action["fuel_mass_fracs"],
                oxyd_mass_fracs=action["oxyd_mass_fracs"],
                atom_ref=action["specfuel"],
                verbose=True,
            )

        if "new_temperature" not in action:
            action["new_temperature"] = None
        if "new_pressure" not in action:
            action["new_pressure"] = None
        if "new_yk" not in action:
            action["new_yk"] = None

        out = alter_state(
            state,
            alpha,
            temp_new=action["new_temperature"],
            press_new=action["new_pressure"],
            y_new=action["new_yk"],
            verbose=True,
        )
        log += "\n" + out

    log += "\n\n Final state"
    log += "\n ==========="

    log += "\n" + state.__repr__()

    return state, log


def gasout_dump_default_input(fname):
    """Dump the default gasout input file"""
    default_gasout_input = pkg_resources.resource_filename(
        __name__, "INPUT/default_gasout_input.yml"
    )
    shutil.copyfile(default_gasout_input, fname)


def _check_action_params(reqd_keys, provided_keys):
    """
    Check gasout action keys

    :param reqd_keys: List of required keys
    :type reqd_keys: list[str]
    :param provided_keys: List of keys in the input file
    :type provided_keys: list[str]

    :returns: Log of the accepted and missing keys
    :rtype: str
    """
    log = str()
    for key in provided_keys:
        if key not in reqd_keys + ["type", "new_temperature", "new_pressure", "new_yk"]:
            log += "\n.   - key " + key + " not accepted"
    if log != "":
        log += "\n.  - Accepted keys"
        log += "\n.      >".join(reqd_keys)

    for key in reqd_keys:
        if key not in provided_keys:
            log += "\n key " + key + " is missing."
    return log


def load_mesh_and_solution(fname, mname):
    """
    Load a mesh and solution into HDF5 objects

    :param fname: Filename of the solution
    :type fname: str
    :param mname: Filename of the mesh
    :type mname: str
    """
    print("Reading solution in ", fname)
    with h5py.File(fname, "r") as fin:
        sol = hdfdict.load(fin, lazy=False)
    print("Reading mesh in ", mname)
    with h5py.File(mname, "r") as fin:
        mesh = hdfdict.load(fin, lazy=False)
    return mesh, sol


def build_data_from_avbp(mesh, sol):
    """
    Build coordinates and a State from an AVBP mesh and solution

    :param mesh: AVBP mesh
    :type mesh: HDF5 object
    :param sol: AVBP solution
    :type sol: HDF5 object

    :returns: Tuple of coordinates and State
    :rtype: tuple
    """
    state = State.from_cons(
        sol["GaseousPhase"]["rho"], sol["GaseousPhase"]["rhoE"], sol["RhoSpecies"]
    )

    x_coor = mesh["Coordinates"]["x"]
    y_coor = mesh["Coordinates"]["y"]
    try:
        z_coor = mesh["Coordinates"]["z"]
        coor = np.stack((x_coor, y_coor, z_coor), axis=-1)
    except KeyError:
        coor = np.stack((x_coor, y_coor), axis=-1)

    return coor, state


def save_data_for_avbp(state, sol, fname):
    """Update the full solution with the state parameters"""
    sol_new = copy.deepcopy(sol)
    sol_new["GaseousPhase"]["rho"] = state.rho
    sol_new["GaseousPhase"]["rhoE"] = state.rho * state.energy
    for spec in sol_new["RhoSpecies"]:
        sol_new["RhoSpecies"][spec] = state.rho * state.mass_fracs[spec]

    try:
        sol_new["Additionals"]["temperature"] = state.temperature
        sol_new["Additionals"]["pressure"] = state.pressure
    except KeyError:
        print("-Additionals- group is not part of the solution.")

    print("Saving solution in ", fname)
    with h5py.File(fname, "w") as fout:
        hdfdict.dump(sol_new, fout)


def gasout_tool(inputfile):
    """Main call"""

    with open(inputfile, "r") as fin:
        in_nob = yaml.load(fin, Loader=yaml.Loader)

    mesh, sol = load_mesh_and_solution(in_nob["inst_solut"], in_nob["mesh"])

    coor, state = build_data_from_avbp(mesh, sol)

    state_new, log = gasout_with_input(coor, state, in_nob)

    print(log)

    save_data_for_avbp(state_new, sol, in_nob["inst_solut_output"])



""" Module testing SpeciesState and MixtureState """

import numpy as np
import pytest

from ms_thermo.mixture_state import SpeciesState, MixtureState

DIM1 = 5

def test_mixture_state():
    """ test of SpeciesAtomic and MixtureAtomic classes """

    ATOMIC_MASS = {
        'C': 12.0107,
        'H': 1.00784,
        'O': 15.999,
        'N': 14.0067}

    def compute_mass_frac(phi, fuel, n_c, n_h):
        m_fuel = max((phi-1) * (n_c*ATOMIC_MASS['C'] + n_h*ATOMIC_MASS['H']), 0.0)
        m_o2 = max((1 - phi) * (n_c+n_h/4) * 2*ATOMIC_MASS['O'], 0.0)
        m_n2 = (n_c+n_h/4) * 3.76 * 2*ATOMIC_MASS['N']
        m_co2 = phi * n_c * (ATOMIC_MASS['C'] + 2*ATOMIC_MASS['O'])
        m_h2o = phi * n_h/2 * (2*ATOMIC_MASS['H'] + ATOMIC_MASS['O'])

        msum = m_fuel + m_o2 + m_n2 + m_co2 + m_h2o
        species_dict = {}
        species_dict[fuel] = m_fuel / msum * np.ones(DIM1)
        species_dict['O2'] = m_o2 / msum * np.ones(DIM1)
        species_dict['N2'] = m_n2 / msum * np.ones(DIM1)
        species_dict['CO2'] = m_co2 / msum * np.ones(DIM1)
        species_dict['H2O'] = m_h2o / msum * np.ones(DIM1)

        return species_dict

    fuel = 'C10H20'
    phi = 1.0
    species_dict = compute_mass_frac(phi, fuel, 10, 20)
    mixture = MixtureState(species_dict, fuel)

    assert len(mixture.species) == 5
    assert isinstance(mixture.species[0], SpeciesState)
    assert mixture.species_name == [fuel, 'O2', 'N2', 'CO2', 'H2O']
    np.testing.assert_allclose(mixture.mixture_fraction, 0.063775, atol=1e-6)
    np.testing.assert_allclose(mixture.equivalence_ratio, phi)
    np.testing.assert_allclose(mixture.far, 0.06812, atol=1e-6)
    np.testing.assert_allclose(mixture.far_st, 0.06812, atol=1e-6)
    np.testing.assert_allclose(mixture.afr, 14.680006, atol=1e-6)

    phi = 0.5
    species_dict = compute_mass_frac(phi, fuel, 10, 20)
    mixture = MixtureState(species_dict, fuel)
    np.testing.assert_allclose(mixture.mixture_fraction, 0.03295, atol=1e-6)
    np.testing.assert_allclose(mixture.equivalence_ratio, phi, atol=2e-4, rtol=4e-4)
    np.testing.assert_allclose(mixture.far, 0.034073, atol=1e-6)
    np.testing.assert_allclose(mixture.far_st, 0.06812, atol=1e-6)
    np.testing.assert_allclose(mixture.afr, 29.348641, atol=1e-6)

    with pytest.raises(NameError):
        mixture.species_by_name('dummy')

    species_dict['N2'] /= 2.0
    with pytest.raises(RuntimeError):
        mixture = MixtureState(species_dict, fuel)



"""
This script calculate mass_fraction of species from a Phi
"""

__all__ = ["yk_from_phi", "phi_from_far"]


MOLAR_MASSES =  {"C": 0.0120107, "H": 0.00100797, "O2": 0.0319988, "N2": 0.0280134}


def far_stochio(c_x:float, h_y:float)-> float:
    """Return the fuel air ratio stoechiometric according to c_x h_y
    
    valid for AIR , not for PURE OXYGEN!!!

    BTW, the example in https://en.wikipedia.org/wiki/Air%E2%80%93fuel_ratio is for pure oxygen...
    """
    mass_mol_fuel = c_x * MOLAR_MASSES["C"] + h_y * MOLAR_MASSES["H"]
    coeff_o2 = c_x + (h_y / 4)
    coef_air=1./0.2314
    afr_sto = coef_air*(coeff_o2 * MOLAR_MASSES["O2"]) / mass_mol_fuel
    far_sto = 1./afr_sto
    return far_sto

def phi_from_far(far:float, c_x:float, h_y:float)-> float:
    """
    *Return phi coefficient with the fuel air ratio coeff + fuel composition*

    :param far: the air-fuel ratio
    :type far: float
    :param c_x: stoechio coeff of Carbone
    :type c_x: float
    :param h_y: stoechio coeff of hydrogene
    :type h_y: float

    :returns:
        - **phi** - Equivalence ratio
    """
    return  far /far_stochio(c_x, h_y)


def far_from_phi(phi:float, c_x:float, h_y:float)-> float:
    """
    *Return fuel air ratio coeff  with the phi coefficient + fuel composition*

    :param phi: eq. ratio coef
    :type far: float
    :param c_x: stoechio coeff of Carbone
    :type c_x: float
    :param h_y: stoechio coeff of hydrogene
    :type h_y: float

    :returns:
        - **far** - Equivalence ratio
    """
    return phi * far_stochio(c_x, h_y)


# use "fuel" here as default. Beware it important for pyavbp.
def yk_from_phi(phi:float, c_x:float, h_y:float, fuel_name:str="fuel")-> float:
    """
    *Return the species mass fractions in a fresh fuel-air mixture*

    :param phi: equivalence ratio
    :type phi: float
    :param c_x: stoechio coeff of Carbone
    :type c_x: float
    :param h_y: stoechio coeff of hydrogene
    :type h_y: float
    :param fuel_name: Name of the fuel
    :type fuel_name: str

    :returns:
        - **y_k** - A dict of mass fractions

    """
    y_k = dict()
    mass_mol_fuel = c_x * MOLAR_MASSES["C"] + h_y * MOLAR_MASSES["H"]
    coeff_o2 = c_x + (h_y / 4)
    y_o2_fuel_sto = (coeff_o2 * MOLAR_MASSES["O2"]) / mass_mol_fuel

    if phi == 0.0:
        y_k[fuel_name] = 0.0
    else:
        y_k[fuel_name] = 1.0 / (
            1.0
            + (1.0 + 3.76 * (MOLAR_MASSES["N2"] / MOLAR_MASSES["O2"]))
            * (y_o2_fuel_sto / phi)
        )
    y_air = 1 - y_k[fuel_name]
    y_k["N2"] = y_air / 1.303794
    y_k["O2"] = 0.303794 * y_k["N2"]

    return y_k




import numpy as np
from ms_thermo.tadia import tadia_table


def test_tadia_table():
	
	burned_t, yburnt = tadia_table(300,101325,1)
	target_value = 2312.990462472097

	np.testing.assert_allclose(target_value, burned_t, rtol=10e-6)

	target_value = [0.7180953905274218, 0, -3.206589963833384e-07, 0.20002461953348916, 0.0818803105980856]
	np.testing.assert_allclose(target_value, [k for k in yburnt.values()], rtol=10e-6)





import pytest
from ms_thermo.yk_from_phi import yk_from_phi, far_from_phi, phi_from_far, far_stochio

def test_far_stochio():
    # test on kerosen
    cx=9.7396
    hy=20.0542
    afr_sr_ref = 14.870368526488049
    #afr_st = 1./far_st_ref
    #assert afr_st==afr_sr_ref  # AFR of Kerozen is around 15
    far_st_ref= 1./afr_sr_ref
    
    far_st=far_stochio(cx,hy)
    assert far_st == far_st_ref

def test_far_from_phi():
    # test from SAE : FAR 102.57 /1000   <> Phi 1.51
    cx=9.7396
    hy=20.0542
    phi_ref=1.51
    far=far_from_phi(phi_ref,cx,hy)
    assert far == 0.10154422180662784

def test_phi_from_far():
    # test from SAE : FAR 102.57 /1000   <> Phi 1.51
    cx=9.7396
    hy=20.0542
    far_ref=0.10257
    phi=phi_from_far(far_ref,cx,hy)
    assert phi == 1.525253699761879 

def test_ykfromphi():
    yf_methane = yk_from_phi(1.0, 1, 4, "CH4")["CH4"]
    yf_c3h8 = yk_from_phi(1.0, 3, 8, "C3H8")["C3H8"]
    yf_c8h18 = yk_from_phi(1.0, 8, 18, "C8H18")["C8H18"]
    yf_h2 = yk_from_phi(1.0, 0, 2, "H2")["H2"]
    pytest.approx(yf_methane, 0.001) == 0.055
    pytest.approx(yf_c3h8, 0.001) == 0.06
    pytest.approx(yf_c8h18, 0.001) == 0.062
    pytest.approx(yf_h2, 0.001) == 0.028
